package edu.washington.cs;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;

import org.json.JSONException;
import org.json.JSONObject;

import com.google.android.maps.GeoPoint;

import edu.washington.cs.listeners.CompassListener;
import edu.washington.cs.listeners.DirectionListener;
import edu.washington.cs.listeners.LocationListener;
import edu.washington.cs.services.CompassService;
import edu.washington.cs.services.DirectionService;
import edu.washington.cs.services.ExchangeLocation;
import android.app.Service;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.location.Location;
import android.os.Binder;
import android.os.Handler;
import android.os.IBinder;
import android.os.Looper;
import android.os.Vibrator;
import android.speech.tts.TextToSpeech;
import android.telephony.TelephonyManager;
import android.util.Log;

/**
 * STORE THE STATE OF THE PROGRAM
 * MANAGE ALL SERVICES IN A MEETUP SISTUATION
 * @author Matt Moyers, Nara Kim
 *
 */

public class MeetUpManager extends Service implements LocationListener, DirectionListener, CompassListener {
	
	private static final String TAG = "MeetUpManager";
	
	//TTS
	public static Vibrator vibrator;
	private TTSManager tts;
	
	//Timeout for server connection in milliseconds
	private final int TIMEOUT = 10000;
	//WebServer URL address
	private final String SERVER_URL = "http://students.washington.edu/mmoyers/meetup/webserver.php?";
	
	private String phoneNumber;
	private String friendID;
	private String name;
	
	private static MeetUpManager self = null;
	
	private final IBinder binder = new MeetUpManagerBinder();
	
	private boolean inMeetUp = false;
	
	private CompassService compassService;
	private DirectionService directionService;
	private ExchangeLocation locationService;
	
	private Intent compassServiceIntent;
	private Intent directionServiceIntent;
	private Intent locationServiceIntent;
	
	private MeetUpServiceConnection serviceConnection = new MeetUpServiceConnection();
	
	
	private RouteStep[] routeStep;
	private String[] gMapCompassDirection = new String[2]; //compass north,east,west,south direction
	private String gMapFirstStep;
	
	private Handler workerHandler;
	private Runnable cancelMeetUpTask;
	private Runnable startMeetUpTask;
	
	private Map<String, String> abbreviationDictionary = new HashMap<String, String>(20);
	
	private RouteStep oldCurrentStep;
	private int currentStepPtr = 0;//initially 0 but when setCurrentStpToNext() is called, incremented.	
	private String totalDuration = "unknown";
	private String totalDistance = "unknown";
	
	
	//State data
	private int state = PRE;
	public static int PRE = 10;
	public static int WAITING = 11;
	public static int ACTIVE = 12;
	
	//String constant for user feedback
	private final String notInMeetUpFeedBack = "You are not currently in a MeetUp.";
	private final String fetchingRouteFeedBack = "Fetching routes from Google Map";
	//TODO: Change to real singleton pattern?
	
	@Override
	public void onCreate() {
	//Find Phone Number
		TelephonyManager telephonyManager = (TelephonyManager)getSystemService(Context.TELEPHONY_SERVICE);
		phoneNumber = telephonyManager.getLine1Number();
		setUpabbreviationDictionary();
		
	//Set Up Update Worker Thread
		Thread thread1 = new Thread() {
			@Override
			public void run() {
				Looper.prepare();
				
				Log.d(TAG, "Creating worker thread.");
				
				workerHandler = new Handler();
				
				Looper.loop();
			}
		};
		thread1.start();
		
	//Set Up Tasks: start, cancel
		cancelMeetUpTask = new Runnable() { 
			public void run() {
				Log.d(TAG, "Cancel Request sent to server.");
				
				try{
					//Make connection to server sending the correct information
					URL url = new URL(SERVER_URL + "action=cancel&id=" + phoneNumber);
					HttpURLConnection con = (HttpURLConnection) url.openConnection();
					
					con.setReadTimeout(TIMEOUT);
					con.setConnectTimeout(TIMEOUT);
					con.setRequestMethod("GET");
					con.setDoInput(true);
					con.connect();
					
					//Receive the data that the server sends us.
					BufferedReader reader = new BufferedReader(new InputStreamReader(con.getInputStream(), "UTF-8"));
					StringBuffer payload = new StringBuffer();
					while(reader.ready()) {
						payload.append(reader.readLine());
					}
					reader.close();
					
					//Parse the data that the server gives us.
					Log.d(TAG, "Cancel Info recieved from the server.");
					
					int status = parseStatusFromJSON(payload.toString());
					if(status != PRE) {
						Log.e(TAG, "Cancel MeetUp action error with server");
					} else {
						setStatus(PRE);
					}
					
				} catch (MalformedURLException e) {
					Log.e(TAG, "Error while attempting to form the url.");
				} catch (IOException e) {
					Log.e(TAG, "CANCEL: Unable to open connection to website.");
					workerHandler.post(cancelMeetUpTask);
				}
			}//end run
		};//end runnable
		
		startMeetUpTask = new Runnable() { 
			public void run() {
				Log.d(TAG, "Start Request Sent to Server.");
				
				try{
					//Make connection to server sending the correct information
					URL url = new URL(SERVER_URL + "action=start&id=" + phoneNumber + "&friend=" + friendID);
					HttpURLConnection con = (HttpURLConnection) url.openConnection();
					
					con.setReadTimeout(TIMEOUT);
					con.setConnectTimeout(TIMEOUT);
					con.setRequestMethod("GET");
					con.setDoInput(true);
					con.connect();
					
					//Receive the data that the server sends us.
					BufferedReader reader = new BufferedReader(new InputStreamReader(con.getInputStream(), "UTF-8"));
					StringBuffer payload = new StringBuffer();
					while(reader.ready()) {
						payload.append(reader.readLine());
					}
					reader.close();
					
					//Parse the data that the server gives us.
					int status = parseStatusFromJSON(payload.toString());
					if(status == WAITING || status == ACTIVE) {
						setStatus(status);
					} else {
						Log.e(TAG, "Start MeetUp action error with server");
					}
					Log.d(TAG, "Start info recieved from the server.");
					
				} catch (MalformedURLException e) {
					Log.e(TAG, "Error while attempting to form the url.");
				} catch (IOException e) {
					Log.e(TAG, "START: Unable to open connection to website.");
					workerHandler.post(startMeetUpTask);
				}
			}//end run
		};//end runnable
		
		vibrator = (Vibrator) getSystemService(Context.VIBRATOR_SERVICE);
		
		//Bind TTSManager
		Intent managerIntent = new Intent(this, TTSManager.class);
		bindService(managerIntent, serviceConnection, Context.BIND_AUTO_CREATE);
		
	}
	
	public static void init(MeetUpManager manager) {
		self = manager;
	}
	
	public static MeetUpManager getInstance() {
		if(self == null) {
			//TODO: throw exception?
			return null;
		} else {
			return self;
		}
	}
	
	public void setStatus(int status) {
		if(status >= 10 && status <= 12) {//VALID STATE
			Log.d(TAG, "SATES: 10->Pre-meetup   11->Waiting for friend   12->Active meet up");
			Log.d(TAG, "SATE CHANGE: " + this.state + " -> " + status);
			
			String speak = "";
			
			if(this.state == PRE && status == WAITING) { //Pre -> Waiting
				speak = "waiting for friend to start meet up";
			} else if(this.state == WAITING && status == ACTIVE) { //Waiting -> Active
				speak = "your friend has started meet up";
			} else if(this.state == ACTIVE && status == WAITING) { //Active -> Waiting
				speak = "your friend has canceled the meet up";
			} else if(this.state == PRE && status == ACTIVE) {//Pre -> Active
				speak = "your friend is ready to meet up";
			} else if(this.state != PRE && status == PRE) { //SOMETHING -> Pre
				speak = "meet up has been canceled";
			}
			this.state = status;
			
			if(tts != null) {
				//tts.speak(speak, TextToSpeech.QUEUE_ADD, null);
				tts.speak(speak, TTSManager.MESSAGE_PRIORITY_IMPORTANT);
			}

			
			
		} else {
			System.out.println("Invalid State.");
		}
	}
	
	public int getStatus() {
		return state;
	}
	
	//Tells if we are in a current meet up with a person
	public boolean areWeInAMeetUp() {
		return inMeetUp;
	}
	
	public boolean areWeInAMeetUp(String id) {
		if(id == null) {
			Log.d(TAG, "id nul?");
			return false;
		}
		return inMeetUp && id.equals(this.friendID);
	}
	
	public String getID() {
		return phoneNumber;
	}
	
	private int parseStatusFromJSON(String json) {
		try {
			JSONObject reply = new JSONObject(json.toString());
			
			return reply.getInt("status");
		
		} catch (JSONException e) {
			Log.e(TAG,  "JSON Error when parsing status.");
		}
		return -1; //ERROR
	}
	
	//Cancels the current meetup
	public void cancelMeetUp() {
		if(inMeetUp) {
			//Send cancel command to server
			workerHandler.post(cancelMeetUpTask);
			
			System.out.println("Shutting down Services");
			
			if(locationService != null) {
				locationService.stop();
				locationService.stopService(locationServiceIntent);
				directionService.stopService(directionServiceIntent);
				compassService.stopService(compassServiceIntent);
			}
			unbindService(serviceConnection);
			
			
			//RESET EVERYTHING
			inMeetUp = false;
			routeStep = null;
		}
	}
	
	//Starts a meet up on a person, if we are in a meetup with a different person that meet up is canceled
	public void startMeetUp(String name, String id) {
		Log.d(TAG, "Starting MEETUP: " + name + " --- " + id);
		
		cancelMeetUp();
		
		//Send start meet up command to the server
		this.friendID = id;
		this.name = name;
		workerHandler.post(startMeetUpTask);
		
		//start up all the services
		
		/*LOCATION*/
		locationServiceIntent = new Intent(this, ExchangeLocation.class);
		startService(locationServiceIntent);
		bindService(locationServiceIntent, serviceConnection, Context.BIND_AUTO_CREATE);		
	    
	    /*DIRECTION*/
		directionServiceIntent = new Intent(this, DirectionService.class);
		//directionServiceIntent.putExtra("travelMode", travelMode);
		startService(directionServiceIntent);
		bindService(directionServiceIntent, serviceConnection, Context.BIND_AUTO_CREATE);
		
		/*COMPASS*/
		compassServiceIntent = new Intent(this, CompassService.class);
		startService(compassServiceIntent);						
		bindService(compassServiceIntent, serviceConnection, Context.BIND_AUTO_CREATE);
		
		inMeetUp = true;
	}
	
	public String getCurrentStep() {
		if(!inMeetUp) {
			return notInMeetUpFeedBack;
		} else if(state == WAITING) {
			return "No steps. Waiting for " + name + " to start meetup.";
		} else if(routeStep == null){
			return fetchingRouteFeedBack;
		} else{
			return routeStep[currentStepPtr].getDescription() +" distance is " 
					+ routeStep[currentStepPtr].getDistance()
					+ " duration is " + routeStep[currentStepPtr].getDuration();
		}		
	}
	
	public String getNextStep(){
		if(!inMeetUp) {
			return notInMeetUpFeedBack;
		} else if(state == WAITING) {
			return "No steps. Waiting for " + name + " to start meetup.";
		} else if(routeStep == null){
			return fetchingRouteFeedBack;
		} else{
			if(routeStep.length > currentStepPtr + 1)
				return routeStep[currentStepPtr+1].getDescription() +" distance is " 
						+ routeStep[currentStepPtr+1].getDistance()
						+ " duration is " + routeStep[currentStepPtr+1].getDuration();
			else 
				return "The current step is the last step to the destiination.";
		}
	}
	
	public String getRouteSummary() {
		if(!inMeetUp) {
			return notInMeetUpFeedBack;
		} else if(state == WAITING) {
			return "No sumary. Waiting for " + name + " to start meetb up.";
		}  else if(routeStep == null){
			return fetchingRouteFeedBack;
		} else{
			StringBuilder description = new StringBuilder();	
			description.append("Total Distance: " + totalDistance 
					+ " Total Duration: " + totalDuration + " ");
			for(int i=0; i < routeStep.length; i++){
				description.append(" step " + (i+1) + " "+ routeStep[i].getDescription()+", ");
			}			
			return description.toString();
		}
	}

	public String getDirectionInFirstStep(){
		String direction="";		
		String readjustMessage="For the best result, please readjust the device and try again.";
		Log.d(TAG, "directionInFirstStep" + gMapFirstStep);
		//define gMapCompassDirection[0] or [1]
		if(inMeetUp){
			if(routeStep!=null){
				getCompassDirectionFromGMap(routeStep[0].getDescription());
				if(gMapFirstStep!=null){
					if(gMapCompassDirection[0]!=null){	
						direction = getFourDirection(gMapFirstStep, gMapCompassDirection[0]);
						if(direction == null)
							return readjustMessage;
						else 
							return direction;
					}else if(gMapCompassDirection[1]!=null){
						direction=getFourDirection(gMapFirstStep, gMapCompassDirection[1]);
						if(direction == null)
							return readjustMessage;
						else
							return direction;
					} else{//can't get Left,Right,String,Back result
						Log.d(TAG, "getDirectionInFirstStep: this statement should never be printed.");
						return readjustMessage;
					}
				} else{
					return fetchingRouteFeedBack;
				}
			} else {
				return fetchingRouteFeedBack;
			}
		} else{
			return notInMeetUpFeedBack;
		}
	}

	@Override
	public void updateCompass(String direction) {
		//Log.d(TAG, "COMPASS UPDATED " + direction);
		if(direction!=null)
			gMapFirstStep = direction;
		else
			gMapFirstStep = null;
		/*if(direction!=null && routeStep != null){
			if(routeStep[0].getDescription()!=null){
				getCompassDirectionFromGMap(routeStep[0].getDescription());
				if(this.gMapCompassDirection[0]!=null){
					userCompassHeading = direction;
				}else if(this.gMapCompassDirection[1]!=null){
					userCompassHeading = direction;
				}
			}
			//stopService(compassServiceIntent);
		}	*/	
	}

	@Override
	public void directionUpdate(RouteStep[] routeStep, String totalDuration, String totalDistance) {	
		Log.d(TAG, "ROUTE UPDATED IN " + TAG);
		if(routeStep == null) {
			Log.e(TAG, "New Route Step is NULL");
			return;
		}
		this.routeStep = routeStep;
		
		flatAbbreviation();
		//Log.d(TAG, "CURRENT DURATION: " + routeStep[currentStepPtr].getDuration());	
		if(oldCurrentStep == null || 
				!this.routeStep[0].isEndLocationEqual(oldCurrentStep.getEndLocation())){
			if(tts != null) {
				//tts.speak("current step has been updated.", TextToSpeech.QUEUE_ADD, null);
				tts.speak("current step has been updated.", TTSManager.MESSAGE_PRIORITY_NORMAL);
			}
			oldCurrentStep = this.routeStep[0];
			currentStepPtr=0;
			this.totalDuration = "unknown";
			this.totalDistance = "unknown";
		}	
		if(totalDuration != null)
			this.totalDuration = totalDuration;
		if(totalDistance != null)			
			this.totalDistance = totalDistance;		
		flatAbbreviationInTotalDuration();
	}

	@Override
	public void locationUpdate(Location source, Location destination) {
		//Log.d(TAG, "LOCATION UPDATED IN " + TAG);
		//System.out.println(source.toString() + "  " + destination.toString());
		
		//Detect if we are together? end of meet up.
		float[] results = new float[3]; //The computed distance is stored in results[0] in meters
		Location.distanceBetween(source.getLatitude(), source.getLongitude(),
				destination.getLatitude(), destination.getLongitude(), results);
		if(results[0] <=  7){ //TODO: Make better feedback.
			//tts.speak("Your friend is within 15 feet.", TextToSpeech.QUEUE_FLUSH, null);
			tts.speak("Your friend is within 15 feet.", TTSManager.MESSAGE_PRIORITY_IMPORTANT);
			cancelMeetUp();
		}

		
		//Not finished yet, process this.
		if(directionService != null) { //as soon as src,dest are updated, start direction
			directionService.start(source, destination);
		}
		
	}
	
	//for user feedback on next step.
	@Override
	public void myLocationUpdate(Location myLocation){
		if(routeStep != null) {
			GeoPoint gp = routeStep[currentStepPtr].getEndLocation();
			float[] results = new float[3]; //The computed distance is stored in results[0] in meters
			Location.distanceBetween(myLocation.getLatitude(), myLocation.getLongitude(),
					gp.getLatitudeE6()/1E6, gp.getLongitudeE6()/1E6, results);
			if(results[0] <= 9 ){ //
				//tts.speak("Your next step is within 30 feet", TextToSpeech.QUEUE_FLUSH, null);
				tts.speak("Your next step is within 30 feet", TTSManager.MESSAGE_PRIORITY_IMPORTANT);
			}
			routeStep[currentStepPtr].setDistance((int)(results[0] * 3.28)
					+ " feet" );
		}
	}
	
	/**
	 * It searches for NEWS and left/right direction and saves them to direction[]
	 * @param routeStep
	 */
	private void getCompassDirectionFromGMap(String routeStep){
		if(routeStep!=null){
			StringTokenizer st = new StringTokenizer(routeStep);
			String step;
			while(st.hasMoreTokens()){
				step=st.nextToken();
				if(step.equals("east") || step.equals("north") || step.equals("south") || step.equals("west")){
					//tts.speak("FORWARD", TextToSpeech.QUEUE_FLUSH, null);
					gMapCompassDirection[0]=step;
					//isDirectionUpdated=true;
					gMapCompassDirection[1]=null;
					return;
				}
				else if(step.equals("left") || step.equals("right")){
					gMapCompassDirection[1]=step;
					//isDirectionUpdated=true;
					gMapCompassDirection[0]=null;
					return;
				}
			}
		}
		gMapCompassDirection[0]=null;
		gMapCompassDirection[1]=null;		
	}
		
	//maps NE -> north east etc..
	private void setUpabbreviationDictionary(){
		abbreviationDictionary.put("ne", "north east");
		abbreviationDictionary.put("se", "south east");
		abbreviationDictionary.put("nw", "north west");
		abbreviationDictionary.put("sw", "south west");
		abbreviationDictionary.put("dr", "drive");
		abbreviationDictionary.put("pl", "place");
		abbreviationDictionary.put("blvd", "boulevard");
		abbreviationDictionary.put("rd", "road");
		abbreviationDictionary.put("ct", "court");
		abbreviationDictionary.put("st", "street");
		abbreviationDictionary.put("ave", "avenue");
		abbreviationDictionary.put("hr", "hour");
		abbreviationDictionary.put("hrs", "hours");
		abbreviationDictionary.put("min", "minute");
		abbreviationDictionary.put("mins", "minutes");
	}
	
	//change description and duration with abbreviations to no abbrivateions using directionDictionary
	private void flatAbbreviation(){
		StringTokenizer tokenDescription, tokenDuration;
		StringBuilder strBuilder = new StringBuilder();
		String oldStr, newStr;
		if(routeStep != null){
			for(int i = 0; i<routeStep.length; i++) {
				//FLAT DESCRIPTION
				tokenDescription = new StringTokenizer(routeStep[i].getDescription());				
				while(tokenDescription.hasMoreTokens()) {					
					oldStr = tokenDescription.nextToken().toLowerCase();
					newStr = abbreviationDictionary.get(oldStr);
					if(newStr == null) {
						strBuilder.append(oldStr);
					} else {
						strBuilder.append(newStr);
					}
					strBuilder.append(" ");
				}
				routeStep[i].setDescription(strBuilder.toString());
				strBuilder = new StringBuilder();
				
				//Log.d(TAG, "DURATION IN FLAT(): " + routeStep[i].getDuration());
				//FLAT DURATION 
				tokenDuration = new StringTokenizer(routeStep[i].getDuration());
				while(tokenDuration.hasMoreTokens()){				
					oldStr = tokenDuration.nextToken();
					newStr = abbreviationDictionary.get(oldStr);
					if(newStr == null) {
						strBuilder.append(oldStr);
					} else {
						strBuilder.append(newStr);
					}
					strBuilder.append(" ");
				}
				routeStep[i].setDuration(strBuilder.toString());
				strBuilder = new StringBuilder();
			}
		}
	}
	//flats string s : intended for flattening total duration/distance
	private void flatAbbreviationInTotalDuration(){
		if(totalDuration != null){
			StringTokenizer tokenizer = new StringTokenizer(totalDuration);
			StringBuilder strBuilder = new StringBuilder();
			String oldStr, newStr;
			while(tokenizer.hasMoreTokens()){				
				oldStr = tokenizer.nextToken();
				newStr = abbreviationDictionary.get(tokenizer);
				if(newStr == null) {
					strBuilder.append(oldStr);
				} else {
					strBuilder.append(newStr);
				}
				strBuilder.append(" ");
			}
			totalDuration = strBuilder.toString();
		}		
	}
	
	
	/**
	 * It returns where the user should face to for the first step. 
	 * @param user
	 * @param gMap
	 * @return
	 */
	private String getFourDirection(String user, String gMap) {
		if(user.equals("n")) {
			if(gMap.equals("north"))
				return "straight";
			else if(gMap.equals("east"))
				return "right";
			else if(gMap.equals("south"))
				return "turn back";
			else if(gMap.equals("west"))
				return "left";
		}
		else if(user.equals("s")){
			if(gMap.equals("south"))
				return "straight";
			else if(gMap.equals("west"))
				return "right";
			else if(gMap.equals("east"))
				return "left";
			else if(gMap.equals("north"))
				return "turn back";					
		}
		else if(user.equals("w")){
			if(gMap.equals("west"))
				return "straight";
			else if(gMap.equals("north"))
				return "right";
			else if(gMap.equals("south"))
				return "left";
			else if(gMap.equals("east"))
				return "turn back";
		}
		else if(user.equals("e")){
			if(gMap.equals("east"))
				return "straight";
			else if(gMap.equals("south"))
				return "right";
			else if(gMap.equals("north"))
				return "left";
			else if(gMap.equals("west"))
				return "turn back";
		}
		return null;
	}	
	
	/*public String getTravelMode(){
		return travelMode;
	}*/
	
	public boolean setTravelMode(String mode){
		if(directionService != null){
			directionService.setTravelMode(mode);
			return true;
		}
		return false;
	}
	
	public class MeetUpServiceConnection implements ServiceConnection {
	 
		@Override
		public void onServiceConnected(ComponentName className, IBinder service){		
			Log.d("MeetUpServiceConnection", className.getShortClassName() + ", " +service.toString());
			
			if(className.getShortClassName().equals(".services.DirectionService")){
				directionService = ((DirectionService.DirectionBinder) service).getService();				
				directionService.addListener(MeetUpManager.this);				
				System.out.println("DirectionService connected");
			} else if(className.getShortClassName().equals(".services.ExchangeLocation")){
				locationService = ((ExchangeLocation.ServiceBinder) service).getService();				
				locationService.addListener(MeetUpManager.this);				
				System.out.println("ExchangeLocation Service connected");
			} else if(className.getShortClassName().equals(".services.CompassService")){
				compassService = ((CompassService.CompassBinder) service).getService();
				compassService.addListener(MeetUpManager.this);
				System.out.println("compassService connected");
			} else if(className.getShortClassName().equals(".TTSManager")){
				Log.d(TAG, "TTSManager Started");
				tts = ((TTSManager.ServiceBinder) service).getService();
			}
		}
		@Override
		public void onServiceDisconnected(ComponentName className){
			if(className.getShortClassName().equals(".DirectionService")) {
				directionService = null;
				stopService(locationServiceIntent);
				System.out.println("DirectionService disconnected");				
			} else if(className.getShortClassName().equals(".ExchangeLocation")){
				locationService = null;
				stopService(directionServiceIntent);
				System.out.println("ExchangeLocation Service disconnected");				
			} else if(className.getShortClassName().equals(".CompassService")){
				compassService = null;
				stopService(compassServiceIntent);
				System.out.println("CompassService disconnected");
			} else if(className.getShortClassName().equals(".TTSManager")){					
				Log.d(TAG, "TTSManager Stopped");
			}
		}
	
	}

	@Override
	public IBinder onBind(Intent intent) {
		return binder;
	}
	
	public class MeetUpManagerBinder extends Binder {
		public MeetUpManager getService() {
			return MeetUpManager.this;
		}
	}
}

