package edu.washington.cs;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Locale;

import android.app.Service;
import android.content.Intent;
import android.os.Binder;
import android.os.IBinder;
import android.speech.tts.TextToSpeech;
import android.speech.tts.TextToSpeech.OnInitListener;
import android.speech.tts.TextToSpeech.OnUtteranceCompletedListener;
import android.util.Log;

/**
 * It manages Text To Speech in the application. 
 * @author Matt Moyers
 *
 */
public class TTSManager extends Service implements OnUtteranceCompletedListener, OnInitListener {
	//DEBUG TAG
	private static final String TAG = "TTSManager";
	//SERVICE
	//Used for hooking into the service and returning a copy of the service
	private final IBinder binder = new ServiceBinder();
	private static TTSManager self;
	
	//MESSAGE LEVELS
	public static final int MESSAGE_PRIORITY_IMPORTANT = 1;
	public static final int MESSAGE_PRIORITY_NORMAL = 0;
	public static final int MESSAGE_PRIORITY_LOW = -1;
	
	private static final int TTS_SPEAKING = 2;
	private static final int TTS_READY = 1;
	private static final int TTS_INIT = 0;
	private static final int TTS_PAUSE = -1;
	
	private static final int TTS_CHECK_CODE = 100;
	
	private ArrayList<Message> list;
	
	//HOW TO GET?
	private TextToSpeech tts;
	
	private HashMap<String, String> map;
	
	//STATE
	private int state;
	
	//CURRENT LEVEL BEING SPOKEN
	private int currentMessageLevel;
	
	@Override
	public void onCreate() {
		list = new ArrayList<Message>();

		tts = new TextToSpeech(this, this);
		
		map = new HashMap<String, String>();
		map.put(TextToSpeech.Engine.KEY_PARAM_UTTERANCE_ID, "utid");
		currentMessageLevel = -99;
		state = TTS_INIT;
	}
	
	public static void init(TTSManager tts) {
		self = tts;
	}
	
	public static TTSManager getInstance() {
		return self;
	}
	
	public void speak(String m, int level) {
		Log.d(TAG, "ADDING : " + m);
		synchronized (list) {
			list.add(new Message(m, level));
		}
		speakNext();
	}
	
	//removes all messages that are equal to or below the level passed
	public void flush(int level) {
		synchronized (list) {
			for(int i = 0; i < list.size(); i++) {
				if(list.get(i).level <= level) {
					Log.d(TAG, "REMOVING: " + list.get(i));
					
					list.remove(i);
				}
			}
		}
		if(currentMessageLevel <= level && state == TTS_SPEAKING) {
			this.stop();
		}
	}
	
	//stops the tts
	public void stop() {
		if(state == TTS_SPEAKING) {
			state = TTS_READY;
			Log.d(TAG, "STOPING TTS");
			tts.stop();
			speakNext();
		}
	}
	
	public void pause() {
		state = TTS_PAUSE;
		tts.stop();
	}
	
	public void resume() {
		if(state == TTS_PAUSE) {
			state = TTS_READY;
			speakNext();
		}
	}
	
	private void speakNext() {
		synchronized (list) {
			if(!list.isEmpty() && state == TTS_READY) {
				Log.d(TAG, "SPEAKING: " + list.get(0).message);
				
				currentMessageLevel = list.get(0).level; 
				tts.speak(list.get(0).message, TextToSpeech.QUEUE_FLUSH, map);//TODO hash map? for params
				try { //TODO: why?
					list.remove(0);
				} catch(IndexOutOfBoundsException e) {
					Log.e(TAG, "Index out of bounds.  Handling...");
				}
				state = TTS_SPEAKING;
			}
		}
	}

	@Override
	public void onUtteranceCompleted(String utteranceId) {
		if(state == TTS_SPEAKING) {
			state = TTS_READY;
		}
		
		speakNext();
	}
	
	/** Function is called when the checkIntent activity returns */
	protected void onActivityResult(int requestCode, int resultCode, Intent data) {
	    if(requestCode == TTS_CHECK_CODE) {
	        if(resultCode == TextToSpeech.Engine.CHECK_VOICE_DATA_PASS) {
	            // success, create the TTS instance
	        	tts = new TextToSpeech(this, this);
	        } else {
	            // missing data, install it
	        	Intent installIntent = new Intent();
	            installIntent.setAction(TextToSpeech.Engine.ACTION_INSTALL_TTS_DATA);
	            startActivity(installIntent);
	        }
	    }
	}
	
	/** Function is called when new TextToSpeechEngine is created */
	@Override
	public void onInit(int status) {
		if(status == TextToSpeech.SUCCESS) {
			//Set up TTS
			tts.setLanguage(Locale.US);
			tts.setOnUtteranceCompletedListener(this);
			tts.setSpeechRate(1.7F);
			
			Log.d(TAG, "Text To Speech Set Up");
			
			state = TTS_READY;
        	speakNext();
		}
	}
	
	private class Message {
		String message;
		int level; //0 expenable 1 nonexpenable
		
		private Message(String m, int l) {
			message = m;
			level = l;
		}
	}
	
	@Override
	public IBinder onBind(Intent intent) {
		return binder;
	}
	
	public class ServiceBinder extends Binder {
		public TTSManager getService() {
			return TTSManager.this;
		}
	}
}
