package edu.washington.cs.services;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.DataFormatException;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.ResponseHandler;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.BasicResponseHandler;
import org.apache.http.impl.client.DefaultHttpClient;
import org.json.JSONArray;
import org.json.JSONObject;
import com.google.android.maps.GeoPoint;

import edu.washington.cs.MeetUpManager;
import edu.washington.cs.RouteStep;
import edu.washington.cs.listeners.DirectionListener;
import android.app.Service;
import android.content.Intent;
import android.location.Location;
import android.os.Binder;
import android.os.Handler;
import android.os.IBinder;
import android.os.Looper;
import android.util.Log;

/**
 * A service to get direction/routes from google maps api.
 * Referred to PhoneWand for processing google maps JSON result.
 * @author Nara Kim
 *
 */
public class DirectionService extends Service {
	
	private final IBinder binder = new DirectionBinder();
	private Location destination;
	private Location myLocation;
	private final String TAG = "DirectionService";		
    private Handler mapHandler;    
    private Thread mapThread;
    private DirectionListener listener;
    private RouteStep[] routeStep;
    private String travelMode = "walking";
    private String totalDuration, totalDistance;
    
    /** Called when the activity is first created. does nothing*/
    @Override
    public void onCreate() {
    	mapThread = new Thread(){
        	@Override
        	public void run(){
        		try {
        			Looper.prepare();
        			mapHandler = new Handler();
        			Log.d(TAG, "Map Thread Started");
        			Looper.loop();
        		} catch (Throwable t){
        			Log.e(TAG, "Map Thread Halted", t);
        		}        		
        	}
       };
       mapThread.start();
    	
    }    
    
    //src: source dest: destination
    public void start(Location src, Location dest){    	
    	myLocation = src;
    	destination = dest;
		callGoogleMap();    	
    }
  
	private String getURL() {	  	       
        // Parse PhoneWandActivity.mCurrentLocation and mDestination into Strings.
        String origin = Double.toString(myLocation.getLatitude())+","+
                        Double.toString(myLocation.getLongitude());
//origin = "47.721069,-122.290766";
        String destination = Double.toString(this.destination.getLatitude())+","+
        					 Double.toString(this.destination.getLongitude());
        
        StringBuilder urlString = new StringBuilder();
        // The URL before parameters.
        urlString.append("http://maps.googleapis.com/maps/api/directions/json?origin=");
        // Get directions from the current location.
        urlString.append(origin);
        // Get directions to the given destination.
        urlString.append("&destination=");
        urlString.append(destination);
        // Get walking directions (as opposed to driving directions), get 
        // only the single best direction, use miles/feet units, and add the 
        // system parameter which is required by Google.
        urlString.append("&mode="+travelMode+"&alternatives=false&units=imperial&sensor=true");
        
        String url = urlString.toString();
        
        Log.d(TAG, url);            
        return url;
	}

    /**
     * Reference: PhoneWand
     * Parse the given polyline into an array of GeoPoints.
     * The algorithm is from http://facstaff.unca.edu/.
     * @param encodedString The polyline String which is to be decoded.
     * @return An array of GeoPoints representing the given polyline.
     */
    private List<GeoPoint> parsePolyLine(String encodedString) throws DataFormatException {
            // Replace double backslashes by single.
            encodedString = encodedString.replace("\\\\", "\\"); //escape char           
            // Decode
            List<GeoPoint> polyline = new ArrayList<GeoPoint>();
            int index = 0, len = encodedString.length();
            int lat = 0, lon = 0;            
            //Log.d(TAG, "start decoding");            
            while (index < len) {
                int b, shift = 0, result = 0;                    
                do {
                    b = encodedString.charAt(index++) - 63;
                    result |= (b & 0x1f) << shift;
                    shift += 5;
                } while (b >= 0x20);                    
                int dlat = ((result & 1) != 0 ? ~(result >> 1) : (result >> 1));
                lat += dlat;                    
                shift = 0;
                result = 0;                    
                if(encodedString.length() > index) {
                    do {
                        b = encodedString.charAt(index++) - 63;
                        result |= (b & 0x1f) << shift;
                        shift += 5;
                    } while (b >= 0x20);
                }
                
                int dlon = ((result & 1) != 0 ? ~(result >> 1) : (result >> 1));
                lon += dlon;                    
                GeoPoint p = new GeoPoint((int)(((double)lat/1E5)*1E6), 
                                 (int)(((double)lon/1E5)*1E6));
                polyline.add(p);                    
                //Log.d(TAG, "p location : lat=" + p.getLatitudeE6() +
                //                	  ", lon=" + p.getLongitudeE6());
            }
            //Log.d(TAG, "end of polyline");            
            return polyline;
    }

	
	//get direction, distance information from google maps api
	private void callGoogleMap(){
		Log.d(TAG, "Google Map Called");
		mapHandler.post(new Runnable(){
			@Override
			public void run(){			    
				HttpClient httpclient = new DefaultHttpClient();
		        HttpGet request = new HttpGet(getURL());		    
		        ResponseHandler<String> handler = new BasicResponseHandler();
		         // Execute request
		        try {		        	 
		        	 //send google map result to parseRoute		     
		        	 routeStep = parseRoute(httpclient.execute(request, handler));
					 listener.directionUpdate(routeStep, totalDuration, totalDistance);
		        	 Log.d(TAG, "Google Map Result is updated");
				} catch (ClientProtocolException e) {					
					 Log.e(TAG, "Google Map: ClientProtocolException", e);
				} catch (IOException e) {
					 Log.e(TAG, "Google Map: IOException", e);
				} catch (DataFormatException e) {
					 Log.e(TAG, "Google Map: DataFormatException", e);
				}		         
		         // Close request
		        httpclient.getConnectionManager().shutdown();			        		      
			}
		});
	}
	
	 /**
	 * Reference: PhoneWand
     * Parses the directions from the given JSON response.  If the routes 
     * array is empty, then null is returned.
     * @param encoded The encoded String from Google Maps.
     * @return An array of Steps parsed from the Google Maps route.
     * @throws DataFormatException if the JSON response is constructed in an 
     *                      unexpected fashion.
     */
	private RouteStep[] parseRoute(String gMapResponse) throws DataFormatException {
        Log.d(TAG, "Start JSON Parsing"); 
        try {
            JSONObject json = new JSONObject(gMapResponse);
            String status = json.getString("status");
            
            if(!status.equals("OK")) { 
            	Log.d(TAG, "google map is used up for today...");
                //throw new Exception("Invalid status: "+status);
            }
            
            JSONArray routes = json.getJSONArray("routes");
            JSONArray legs = routes.getJSONObject(0).getJSONArray("legs");
            JSONArray steps = legs.getJSONObject(0).getJSONArray("steps");
            
            totalDistance = legs.getJSONObject(0).getJSONObject("distance").getString("text");
            Log.d(TAG, "TOTAL DISTANCE IS " + totalDistance);
            totalDuration = legs.getJSONObject(0).getJSONObject("duration").getString("text");
            
            int length = steps.length();            
            RouteStep[] stepArray = new RouteStep[length];
            
            JSONObject jsonStep, location;
            GeoPoint currentStart, currentEnd;
            List<GeoPoint> currentPolyline;
            String currentDuration, currentDistance, currentDescription;
            int lat, lon;
            
            // Loop over the steps, creating a new Step object for each 
            // and adding it to the mSteps array.
            for(int i = 0; i < length; i++) {     
                // Start location.
                jsonStep = steps.getJSONObject(i);
                location = jsonStep.getJSONObject("start_location");
                lat = (int)(location.getDouble("lat"));
                lon = (int)(location.getDouble("lng"));
                currentStart = new GeoPoint(lat, lon);
                // End location.
                location = jsonStep.getJSONObject("end_location");
                lat = (int)(location.getDouble("lat")*1E6);
                lon = (int)(location.getDouble("lng")*1E6);
                currentEnd = new GeoPoint(lat, lon);
                // Distance and duration.
                currentDuration = jsonStep.getJSONObject("duration").getString("text");
                
                currentDistance = jsonStep.getJSONObject("distance").getString("text");
                // Textual description.
                currentDescription = jsonStep.getString("html_instructions");
                currentDescription = currentDescription.replaceAll("<div", " <div");
                currentDescription = currentDescription.replaceAll("<(.|\n)*?>", "");
                // Set of GeoPoints detailing the step's path.
                currentPolyline = parsePolyLine(jsonStep.getJSONObject("polyline").getString("points"));
                
                // Create and add new Step.
                stepArray[i] = new RouteStep(currentStart, currentEnd, currentPolyline, 
                		currentDistance, currentDuration, currentDescription);
            }
            
            Log.d(TAG, "End JSON Parsing");
            return stepArray;
                
        } catch(Exception e) {
            Log.e(TAG, "JSON Parsing Error", e);
        }
		return null;
    }

	public void setTravelMode(String mode){
		travelMode = mode;
		if(MeetUpManager.getInstance().getStatus() == MeetUpManager.ACTIVE) {
			callGoogleMap();
		}
	}
	
	@Override
	public IBinder onBind(Intent intent) {
		return binder;
	}
	
	public void addListener(DirectionListener listener){
		this.listener = listener;
	}
	
	/**
	 * Inner class to bind MyLocationService to main activity.
	 * @author Nara Kim
	 *
	 */
	public class DirectionBinder extends Binder {
		public DirectionService getService(){
			return DirectionService.this;
		}
	}

}
