package edu.washington.cs.services;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import edu.washington.cs.MeetUpManager;

import android.app.Service;
import android.content.Context;
import android.content.Intent;
import android.location.Criteria;
import android.location.Location;
import android.location.LocationListener;
import android.location.LocationManager;
import android.os.Binder;
import android.os.Bundle;
import android.os.Handler;
import android.os.IBinder;
import android.os.Looper;
import android.telephony.TelephonyManager;
import android.util.Log;

/**
 * 
 * This class is to send user's current location to server 
 * and receive friend's current location.
 * 
 * Event driven
 * 
 *  @author Matt Moyers
 *  @version 1.00
 * 
 * 
 **************************************
 *			EXAMPLE
 **************************************
 * 
 * ----How to implement this class as a service----
 * 
 * ----SomeClass.java----
 * 
 * public class SomeClass implements ServiceConnection {
 * 
 *  //variable to hold a link to the service we are starting
 *  private ExchangeLocation exchangeLocation;
 *  //hold the intent used for stopping the service
 *  private Intent exchangeLocationIntent;
 *  
 *  //call to bind this class to the service
 *  public void createExchangeLocation() {
 *    //Start the service
 *    startService(exchangeLocationIntent);
 *    //This binds this class to the service
 *    bindService(new Intent(this, ExchangeLocation.class), this, Context.BIND_AUTO_CREATE);
 *  }
 *  
 *  //call to stop the service
 *  public void stopExchangeLocation() {
 *    stopService(exchangeLocationIntent);
 *  }
 * 
 *  //Called when the service binding is complete and the service is started
 * 	@Override
 *	public void onServiceConnected(ComponentName name, IBinder service) {
 *		//Once connected to the service get a link to the service
 *		exchangeLocation = ((ExchangeLocation.ServiceBinder) service).getService();
 *	}
 *
 * }
 * 
 * 
 * 
 * ----Android Manifest----
 * ----In the manifest add this----
 * 
 * <application>
 * .
 * .
 * .
 * 	<service android:name=".ExchangeLocation"></service>
 *       
 *  <receiver android:name=".ServerAlarmReciever"></receiver>
 * </application>
 * 
 * 
 * 
 *
 *
 */

public class ExchangeLocation extends Service implements LocationListener {
	//Tag for log
	private static final String TAG = "ExchangeLocation";
	//Speed in millisecond that we refresh the location
	private static final int UPDATE_SPEED = 5000;
	//Timeout for server connection in milliseconds
	private static final int TIMEOUT = 3000;
	//WebServer URL address
	private static final String SERVER_URL = "http://students.washington.edu/mmoyers/meetup/webserver.php";
	//Used for hooking into the service and returning a copy of the service
	private final IBinder binder = new ServiceBinder();
	//For GPS shutdown
	private LocationManager loc;
	
	
	//Source and Desintation Locations
	private Location myLocation;
	private Location serverLocation;
	
	//Handlers for each thread server communication and listener communication
	private Handler serverHandler;
	private Handler listenerHandler;
	
	//People who are listening to me for location updates
	private edu.washington.cs.listeners.LocationListener listener;
	
	//Stored task to update with server
	private Runnable updateServerTask;
	
	
	@Override
	public void onCreate() {	
		Log.d(TAG, "CREATING EXCHANGELOCATION SERVICE");
		
		//Set Up Update Listener Thread
		Thread thread1 = new Thread() {
			@Override
			public void run() {
				Looper.prepare();
				
				Log.d(TAG, "Creating listner thread.");
				
				listenerHandler = new Handler();
				
				Looper.loop();
			}
		};
		thread1.start();
		
		//Set up server communcation thread
		Thread thread2 = new Thread() {
			@Override
			public void run() {
				if(serverHandler != null) { //Illegal call to this method was detected
					Log.d(TAG, "Illegal call to server communication thread detected. Handling...");
					return;
					//throw new IllegalAccessException();
				}
				
				Looper.prepare();
				
				Log.d(TAG, "Creating server communication thread.");
				
				serverHandler = new Handler();
				
				Looper.loop();
			}
		};
		thread2.start();
		
		updateServerTask = new Runnable() { 
			public void run() {
				Log.d(TAG, "Updating info with the server.");
				
				try{			
					//Make connection to server sending the correct information
					URL url = new URL(SERVER_URL + "?action=update&id=" + MeetUpManager.getInstance().getID() + parseLocationToString());
					HttpURLConnection con = (HttpURLConnection) url.openConnection();
					
					con.setReadTimeout(TIMEOUT);
					con.setConnectTimeout(TIMEOUT);
					con.setRequestMethod("GET");
					con.setDoInput(true);
					con.connect();
					
					//Receive the data that the server sends us.
					BufferedReader reader = new BufferedReader(new InputStreamReader(con.getInputStream(), "UTF-8"));
					StringBuffer payload = new StringBuffer();
					while(reader.ready()) {
						payload.append(reader.readLine());
					}
					reader.close();
					Log.d(TAG, "Info recieved from the server.");
					
					//Parse the data that the server gives us.
					synchronized(serverLocation) {
						Location loc = createLocationFromJSON(payload.toString());
						if(loc != null) { //only update if new info is found
							serverLocation = loc;
							notifyListeners();
						}
						
					}
					
					//Log.d(TAG, "Update destination location successful.");
					
				} catch (MalformedURLException e) {
					Log.e(TAG, "Error while attempting to form the url.");
				} catch (IOException e) {
					Log.e(TAG, "Unable to open connection to website.");
					serverHandler.post(updateServerTask);
					return;
				}
				
				//Loop the update
				serverHandler.postDelayed(updateServerTask, UPDATE_SPEED);
			}//end run
		};//end runnable
		
		//Set up GPS
		loc = (LocationManager) getSystemService(LOCATION_SERVICE);
		Log.d(TAG, "GPS set up");
		
		//Set up location
		myLocation = loc.getLastKnownLocation(LocationManager.GPS_PROVIDER);
		if(myLocation == null) {
			myLocation = new Location("Default");
		}
		serverLocation = new Location("Default");
	}
	
	//Turns Location into a String
	private String parseLocationToString() {
		synchronized(myLocation) {
			return "&acc=" + myLocation.getAccuracy() + 
				"&alt=" + myLocation.getAltitude() + 
				"&bear=" + myLocation.getBearing() + 
				"&lat=" + myLocation.getLatitude() + 
				"&long=" + myLocation.getLongitude() + 
				"&speed=" + myLocation.getSpeed();
		}
	}
	
	@Override
	public int onStartCommand(Intent intent, int flag, int startId) {
		//Start the updates.
		Log.d(TAG, "Starting the exchange of location data with the server.");
		try {
			Thread.sleep(100);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		Log.d(TAG, "Hooking into GPS");
		//Criteria criteria = new Criteria();
		//criteria.setAccuracy(accuracy);
		//String provider = loc.getBestProvider(criteria, true);
		//loc.requestLocationUpdates(provider, 0, 0, this);
		loc.requestLocationUpdates(LocationManager.GPS_PROVIDER, 0, 0, this);
		
		serverHandler.post(updateServerTask);
		
		return START_STICKY;
	}
	
	@Override
	public void onDestroy() {
		serverHandler.getLooper().quit();
		listenerHandler.getLooper().quit();
		loc.removeUpdates(this);
		Log.d(TAG, "Destroying ExchangeLocation.");
	}
	
	@Override
	public IBinder onBind(Intent intent) {
		return binder;
	}
	
	public void addListener(edu.washington.cs.listeners.LocationListener serverListener) {
		listener = serverListener;
		//.d(TAG, "Adding Listener.");
	}
	
	private void notifyListeners() {
		listenerHandler.post(new Runnable() {
			@Override
			public void run() {
				if(listener != null) {
					//.d(TAG, "Notifying Listeners.");
					synchronized(serverLocation) {
						listener.locationUpdate(myLocation, serverLocation);
					}
				} else {
					Log.d(TAG, "No Listeners to Notify.");
					
				}
			}
		});
	}
	
	//Converts an XML file into a Location Object
	private Location createLocationFromJSON(String json) {
		
		Location location = null;
		
		try {
			JSONObject reply = new JSONObject(json.toString());
	
			int status = reply.getInt("status");
			System.out.println("Server:" + status + " Mine:" + MeetUpManager.getInstance().getStatus());
			if(status != MeetUpManager.getInstance().getStatus()) {
				MeetUpManager.getInstance().setStatus(status);
			}
			
			if(status == MeetUpManager.ACTIVE) {
				JSONArray people = reply.getJSONArray("person");
				for(int i = 0; i < people.length(); i++) {
					JSONObject person = people.getJSONObject(i);
					location = new Location(String.valueOf(person.getLong("id")));
					location.setAccuracy(Float.parseFloat(person.getString("accuracy")));
					location.setAltitude(person.getDouble("altitude"));
					location.setBearing(Float.parseFloat(person.getString("bearing")));
					location.setLatitude(person.getDouble("latitude"));
					location.setLongitude(person.getDouble("longitude"));
					location.setSpeed(Float.parseFloat(person.getString("speed")));
				}
			}
		
		} catch (JSONException e) {
			Log.e(TAG, "JSONException.", e);
		}
		
		return location;
	}
	
	public void stop() {
		serverHandler.removeCallbacks(updateServerTask);
		loc.removeUpdates(this);
		Log.d(TAG, "Stopping the exchange of location data with the server.");
	}
	
	public void onLocationChanged(Location location) {
		//Update last known location using thread locking to ensure thread safety
		synchronized(myLocation) {
			myLocation = location;
		}
		if(listener != null) {
			listener.myLocationUpdate(location);
		}
		
		
		
		//Log.d(TAG, "Location from GPS has been updated");
	}

	public void onProviderDisabled(String provider) { }

	public void onProviderEnabled(String provider) { }

	public void onStatusChanged(String provider, int status, Bundle extras) { }
	
	public class ServiceBinder extends Binder {
		public ExchangeLocation getService() {
			return ExchangeLocation.this;
		}
	}

}